/*
 * xbloody.c
 *	Copyright (C) 1994 by K.Emoto
 */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>

typedef struct _RootWinClassRec*	RootWinWidgetClass;
typedef struct _RootWinRec*	RootWinWidget;
extern WidgetClass rootWinWidgetClass;

typedef struct {
    int empty;
} RootWinwClassPart;

typedef struct _RootWinwClassRec {
    CoreClassPart	core_class;
    RootWinwClassPart	root_class;
} RootWinwClassRec;

extern RootWinwClassRec rootClassRec;

typedef struct {
    /* resources */
    char* resource;
    /* private state */
} RootWinwPart;

typedef struct _RootWinwRec {
    CorePart	core;
    RootWinwPart	root;
} RootWinwRec;


static void Realize();

RootWinwClassRec rootWinClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"RootWinw",
    /* widget_size		*/	sizeof(RootWinwRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	NULL,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	NULL,
    /* num_resources		*/	0,
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	NULL,
    /* set_values		*/	NULL,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* rootWin fields */
    /* empty			*/	0
  }
};

WidgetClass rootWinWidgetClass = (WidgetClass)&rootWinClassRec;

static void Realize(w, value_mask, attributes)
Widget	w;
XtValueMask *value_mask;
XSetWindowAttributes *attributes;
{
    w->core.window = RootWindowOfScreen(w->core.screen);
}


void	GrabRootWin(rootW, cursor, eh, client_data)
Widget	rootW;
Cursor	cursor;
XtEventHandler	eh;
XtPointer client_data;
{
	XtGrabPointer(rootW, False,
		Button1MotionMask|ButtonPressMask|ButtonReleaseMask,
		GrabModeAsync, GrabModeAsync, None,
		cursor, CurrentTime);
	XtAddRawEventHandler(rootW,
		Button1MotionMask|ButtonPressMask|ButtonReleaseMask,
		True, (XtEventHandler)eh, client_data);
}

void	UngrabRootWin(rootW, cursor, eh, client_data)
Widget	rootW;
Cursor	cursor;
XtEventHandler	eh;
XtPointer client_data;
{
	XtUngrabPointer(rootW, CurrentTime);
	XtRemoveRawEventHandler(rootW,
		PointerMotionMask|ButtonPressMask|ButtonReleaseMask,
		True, (XtEventHandler)eh, client_data);

}


#include "knife.xbm"
#include "knife_mask.xbm"

Widget  toplevel;
XtAppContext    app;
Display *dpy;
char	*version;
Widget	RootW;
Cursor	cursor;
GC		gc, bgc;
XPoint	startPoint, endPoint;
#define	NUMCOLORS	3
Pixel	pix[NUMCOLORS];
int	junk;
unsigned int width, height;

static void DragEH();
int	interval;
String	color;

static XtResource application_rsc[] = {
    {"interval", "Interval", XtRInt, sizeof (int),
    (Cardinal)&interval, XtRImmediate, (caddr_t)10 },
    {"color", "Color", XtRString, sizeof (String),
    (Cardinal)&color, XtRString, (caddr_t)"red" },
};

static  XrmOptionDescRec    options[] = {
    { "-interval",   "interval",  XrmoptionSepArg,    NULL    },
    { "-color",   "color",  XrmoptionSepArg,    NULL    },
};

usage()
{
	fprintf(stderr, "Usage: xbloody [options]\n");
	fprintf(stderr, "        -interval <interval>: draw interval by msec\n");
	fprintf(stderr, "        -color <color>: blood color name\n");
	fprintf(stderr, "      exit by <Button3 Release>\n");
}

main(argc, argv)
char    **argv;
{
	Arg     args[10];
	int     n;

	fprintf(stderr, "xbloody version 1.00 by Katsuhiro Emoto\n");

	toplevel = XtAppInitialize(&app, "XBloody",
					options, XtNumber(options), &argc, argv,
					NULL, NULL, 0);
	if (argc > 1) {
		usage();
		exit(1);
	}
	dpy = XtDisplay(toplevel);
	XtGetApplicationResources(toplevel, (caddr_t)NULL, application_rsc,
			XtNumber(application_rsc), NULL, 0);
	setup();
	XtAppMainLoop(app);
}

setup()
{
	{
		Pixmap	source, mask;
		char	buf[1024];
		int x_hot, y_hot, dumm2;
		Dimension dumm1;
		int width, height;
		XColor foreground_color, background_color, dumm;

		XLookupColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)),
			"black", &foreground_color, &dumm);
		XLookupColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)),
			"white", &background_color, &dumm);

		source = XCreateBitmapFromData(dpy,
			RootWindow(dpy, DefaultScreen(dpy)),
			knife_bits, knife_width,
			knife_height);
		mask = XCreateBitmapFromData(dpy,
			RootWindow(dpy, DefaultScreen(dpy)),
			knife_mask_bits, knife_mask_width,
			knife_mask_height);
 
		if (source && mask) {
			cursor = XCreatePixmapCursor(dpy, source, mask,
				&foreground_color, &background_color, 2, 2);
		}
		if (source) {
			XFreePixmap(dpy, source);
		}
		if (mask) {
			XFreePixmap(dpy, mask);
		}

		if (!cursor) {
			cursor = XCreateFontCursor(dpy, XC_pirate);
		}
	}

    RootW = XtCreateManagedWidget("root",
                rootWinWidgetClass,
                toplevel, NULL, 0 );
    XtRealizeWidget(RootW);

    {
		XGCValues   GCValues;
		unsigned    long    ValueMask;

		ValueMask = GCFunction|GCForeground|GCPlaneMask|GCLineWidth| GCLineStyle
		;
		GCValues.function   = GXxor;
		GCValues.foreground = 0xffffffff;
		GCValues.plane_mask = 0xffffffff;
		GCValues.line_width = 1;
		GCValues.line_style = LineSolid;
		gc = XCreateGC(dpy, XtWindow(RootW), ValueMask, &GCValues);
		XSetSubwindowMode(dpy, gc, IncludeInferiors);
    }
	{
		int	i;
		XColor	xc;
		XColor	bc, dumm;

		if (XLookupColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)),
			color, &bc, &dumm) == 0) {
			fprintf(stderr, "color %s not found\n", color);
			exit(1);
		}
#if 0
printf("rbg = %d,%d,%d\n", bc.red, bc.green, bc.blue);
#endif
		for (i = 0; i < NUMCOLORS; i++) {
			int	val;
			val = i * (bc.red >> 8) / (NUMCOLORS-1);
			xc.red = (val << 8) | val;
			val = i * (bc.green >> 8) / (NUMCOLORS-1);
			xc.green = (val << 8) | val;
			val = i * (bc.blue >> 8) / (NUMCOLORS-1);
			xc.blue = (val << 8) | val;
			xc.pixel = 0;
			XAllocColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), &xc);
			pix[i] = xc.pixel;
#if 0
printf("pixel = %x\n", xc.pixel);
#endif
		}
	}
    {
		XGCValues   GCValues;
		unsigned    long    ValueMask;

		ValueMask = GCFunction|GCForeground|GCPlaneMask|GCLineWidth| GCLineStyle
		;
		GCValues.function   = GXcopy;
		GCValues.foreground = 0x00ff0000;
		GCValues.plane_mask = 0xffffffff;
		GCValues.line_width = 1;
		GCValues.line_style = LineSolid;
		bgc = XCreateGC(dpy, XtWindow(RootW), ValueMask, &GCValues);
		XSetSubwindowMode(dpy, bgc, IncludeInferiors);
    }

	GrabRootWin(RootW, cursor, DragEH, NULL);
	XGetGeometry(dpy, XtWindow(RootW), (Window *)&junk,
		(int *)&junk, (int *)&junk,
		(unsigned int *)&width, (unsigned int *)&height,
		(unsigned int *)&junk, (unsigned int *)&junk);
}

static void
DragEH(w, rec, event, continue_to_dispatch)
Widget w;
void	*rec;
XEvent *event;
Boolean *continue_to_dispatch;
{
	switch (event->type) {
	case ButtonRelease:
		if ( event->xbutton.state & Button3Mask ) {
			UngrabRootWin(RootW, cursor, DragEH, NULL);
			XtDestroyWidget(RootW);
			exit(0);
		}
		DrawBlood();
		break;
	case ButtonPress:
		startPoint.x = event->xbutton.x_root;
		startPoint.y = event->xbutton.y_root;
		endPoint.x = event->xbutton.x_root;
		endPoint.y = event->xbutton.y_root;
		break;
	case MotionNotify:
		XDrawLine(dpy, XtWindow(RootW), gc,
			startPoint.x, startPoint.y, endPoint.x, endPoint.y);
		endPoint.x = event->xmotion.x_root;
		endPoint.y = event->xmotion.y_root;
		XDrawLine(dpy, XtWindow(RootW), gc,
			startPoint.x, startPoint.y, endPoint.x, endPoint.y);
		break;
	}
}

int      wSine[] = {
    0,  3,  6,  9, 12, 16, 19, 22, 24, 27, 30, 33, 36, 38,
    41, 43, 45, 47, 49, 51, 53, 55, 56, 58, 59, 60, 61, 62,
    63, 63, 64, 64, 64, 64, 64, 63, 63, 62, 61, 60, 59, 58,
    56, 55, 53, 51, 49, 47, 45, 43, 41, 38, 36, 33, 30, 27,
    24, 22, 19, 16, 12,  9,  6,  3,  0};

typedef struct {
    int		y;
    int		cur;
	int		rate;
} INFO;

typedef	struct	{
	INFO	*info;
	int		nPoint;
	int		count;
	int		direct;
	XPoint	startPoint, endPoint;
	int		nWidth;
	int		iw;
	int		x,y;
} REC;

#define	min(a,b)	(((a)>(b))?(b):(a))

DrawBloodSegment(rec)
REC	*rec;
{
	int	nWidth;
	int	x, y;
	int		nPoint;

	nPoint = rec->nPoint;
	if (rec->iw == 0) {
		rec->nWidth = (rand() % min(64, nPoint)) + 1;
		rec->x = rand() % (nPoint + 1 - rec->nWidth);
		rec->y = (rand() % 64) + 1;
		rec->count++;
	}
	nWidth = rec->nWidth;
	x = rec->x;
	y = rec->y;
	{
		int	newy;
		int	pos;
		INFO	*ip;

		pos = x + rec->iw;
		ip = &rec->info[pos];
		newy = ip->cur;
		if (newy > height) {
		}
		else {
			newy += wSine[rec->iw * nWidth / 64] * y / 256;
			XDrawLine(dpy, XtWindow(RootW), bgc,
					rec->startPoint.x + pos * rec->direct, ip->cur, 
					rec->startPoint.x + pos * rec->direct, newy);
			ip->cur = newy;
		}
		rec->iw++;
		if (rec->iw >= rec->nWidth) {
			rec->iw = 0;
		}
	}
}
DrawBloodLine(rec)
REC	*rec;
{
	rec->iw = 0;
	DrawBloodSegment(rec);
	for (; rec->iw ; ) {
		DrawBloodSegment(rec);
	}
	XFlush(dpy);
}

TimerCB(rec, id)
REC	*rec;
XtIntervalId	*id;
{
	DrawBloodLine(rec);
	if (rec->count < 500) {
		XtAppAddTimeOut(app, interval, (XtTimerCallbackProc)TimerCB, (XtPointer)rec);
	}
	else {
		XtFree((XtPointer)rec->info);
		XtFree((XtPointer)rec);
	}
}

DrawBlood()
{
	int	i, j;
	int	x, y;
	int	nPoint;
	INFO	*info;
	int	direct;
	double	dx;
	REC	*rec;

	if (startPoint.x == endPoint.x) {
		return 0;
	}
	if (endPoint.x > startPoint.x) {
		direct = 1;
		nPoint = endPoint.x - startPoint.x;
	}
	else {
		direct = -1;
		nPoint = startPoint.x - endPoint.x;
	}
	rec = (REC *)XtCalloc(1, sizeof(REC));
	info = (INFO *)XtCalloc(nPoint, sizeof(INFO));
	rec->info = info;
	rec->startPoint = startPoint;
	rec->endPoint = endPoint;
	rec->nPoint = nPoint;
	rec->direct = direct;
	rec->count = 0;
	rec->iw = 0;

	dx = ((double)(endPoint.y - startPoint.y))/
			((double)(endPoint.x - startPoint.x));
	dx *= direct;
	for (i = 0; i < nPoint; i++) {
		info[i].y = startPoint.y + (i * dx);
		info[i].cur = info[i].y;
	}

	for (i = 0; i < NUMCOLORS; i++) {
		XSetForeground(dpy, bgc, pix[i]);
		XDrawLine(dpy, XtWindow(RootW), bgc,
				startPoint.x, startPoint.y - i,
				endPoint.x, endPoint.y - i);
		XDrawLine(dpy, XtWindow(RootW), bgc,
				startPoint.x, startPoint.y + i,
				endPoint.x, endPoint.y + i);
	}

	XtAppAddTimeOut(app, interval, (XtTimerCallbackProc)TimerCB, (XtPointer)rec);
}

